// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

/******************************************************************

DO NOT MODIFY. AUTOGENERATED FILE.
This file is generated using the logic from <root>/src/scripts/genEventPipe.py

******************************************************************/

#include <eventpipe/ep-rt-config.h>
#ifdef ENABLE_PERFTRACING
#include <eventpipe/ep.h>
#include <eventpipe/ep-event.h>
#include "clreventpipewriteevents.h"

#ifndef W
#define W(str) L##str
#endif
#ifndef SL
#define SL(str) str
#endif
#ifndef ERROR_SUCCESS
#define ERROR_SUCCESS 0L
#endif
#ifndef ERROR_WRITE_FAULT
#define ERROR_WRITE_FAULT 29L
#endif

/*
 * Forward declares of functions.
 */

bool
resize_buffer (
    uint8_t **buffer,
    size_t *size,
    size_t current_size,
    size_t new_size,
    bool *fixed_buffer);

bool
write_buffer (
    const uint8_t *value,
    size_t value_size,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer);

bool
write_buffer_string_utf8_t (
    const ep_char8_t *value,
    size_t value_len,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer);

bool
write_buffer_string_utf8_to_utf16_t (
    const ep_char8_t *value,
    size_t value_len,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer);

static
inline
bool
write_buffer_guid_t (
    const uint8_t *value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    return write_buffer (value, EP_GUID_SIZE, buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_uint8_t (
    uint8_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    return write_buffer ((const uint8_t *)&value, sizeof (uint8_t), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_uint16_t (
    uint16_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    value = ep_rt_val_uint16_t (value);
    return write_buffer ((const uint8_t *)&value, sizeof (uint16_t), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_uint32_t (
    uint32_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    value = ep_rt_val_uint32_t (value);
    return write_buffer ((const uint8_t *)&value, sizeof (uint32_t), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_int32_t (
    int32_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    value = ep_rt_val_int32_t (value);
    return write_buffer ((const uint8_t *)&value, sizeof (int32_t), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_uint64_t (
    uint64_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    value = ep_rt_val_uint64_t (value);
    return write_buffer ((const uint8_t *)&value, sizeof (uint64_t), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_int64_t (
    int64_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    value = ep_rt_val_int64_t (value);
    return write_buffer ((const uint8_t *)&value, sizeof (int64_t), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_double_t (
    double value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
#if BIGENDIAN
    uint64_t value_as_uint64_t;
    memcpy (&value_as_uint64_t, &value, sizeof (uint64_t));
    value_as_uint64_t = ep_rt_val_uint64_t (value_as_uint64_t);
    memcpy (&value, &value_as_uint64_t, sizeof (uint64_t));
#endif
    return write_buffer ((const uint8_t *)&value, sizeof (double), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_bool_t (
    bool value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    return write_buffer_int32_t (value, buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_uintptr_t (
    uintptr_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    value = ep_rt_val_uintptr_t (value);
    return write_buffer ((const uint8_t *)&value, sizeof (uintptr_t), buffer, offset, size, fixed_buffer);
}

static
inline
EventPipeEvent *
provider_add_event (
    EventPipeProvider *provider,
    uint32_t event_id,
    uint64_t keywords,
    uint32_t event_version,
    EventPipeEventLevel level,
    bool need_stack)
{
    return ep_provider_add_event (provider, event_id, keywords, event_version, level, need_stack, NULL, 0);
}

static
inline
EventPipeProvider *
create_provider (
    const wchar_t *provider_name,
    EventPipeCallback callback_func)
{
    ep_char8_t *provider_name_utf8 = NULL;

#if WCHAR_MAX == 0xFFFF
    provider_name_utf8 = g_utf16_to_utf8 ((const gunichar2 *)provider_name, -1, NULL, NULL, NULL);
#else
    gunichar2 *provider_name_utf16 = g_ucs4_to_utf16 ((const gunichar *)provider_name, -1, NULL, NULL, NULL);
    provider_name_utf8 = g_utf16_to_utf8 (provider_name_utf16, -1, NULL, NULL, NULL);
    g_free (provider_name_utf16);
#endif

    ep_return_null_if_nok (provider_name_utf8 != NULL);

    EventPipeProvider *provider = ep_create_provider (provider_name_utf8, callback_func, NULL);

    g_free (provider_name_utf8);
    return provider;
}

const wchar_t* DotNETRuntimeMonoProfilerName = W("Microsoft-DotNETRuntimeMonoProfiler");
EventPipeProvider *EventPipeProviderDotNETRuntimeMonoProfiler = NULL;
EventPipeEvent *EventPipeEventMonoProfilerContextLoaded = NULL;
EventPipeEvent *EventPipeEventMonoProfilerContextUnloaded = NULL;
EventPipeEvent *EventPipeEventMonoProfilerAppDomainLoading = NULL;
EventPipeEvent *EventPipeEventMonoProfilerAppDomainLoaded = NULL;
EventPipeEvent *EventPipeEventMonoProfilerAppDomainUnloading = NULL;
EventPipeEvent *EventPipeEventMonoProfilerAppDomainUnloaded = NULL;
EventPipeEvent *EventPipeEventMonoProfilerAppDomainName = NULL;
EventPipeEvent *EventPipeEventMonoProfilerJitBegin = NULL;
EventPipeEvent *EventPipeEventMonoProfilerJitFailed = NULL;
EventPipeEvent *EventPipeEventMonoProfilerJitDone = NULL;
EventPipeEvent *EventPipeEventMonoProfilerJitDone_V1 = NULL;
EventPipeEvent *EventPipeEventMonoProfilerJitChunkCreated = NULL;
EventPipeEvent *EventPipeEventMonoProfilerJitChunkDestroyed = NULL;
EventPipeEvent *EventPipeEventMonoProfilerJitCodeBuffer = NULL;
EventPipeEvent *EventPipeEventMonoProfilerClassLoading = NULL;
EventPipeEvent *EventPipeEventMonoProfilerClassFailed = NULL;
EventPipeEvent *EventPipeEventMonoProfilerClassLoaded = NULL;
EventPipeEvent *EventPipeEventMonoProfilerClassLoaded_V1 = NULL;
EventPipeEvent *EventPipeEventMonoProfilerVTableLoading = NULL;
EventPipeEvent *EventPipeEventMonoProfilerVTableFailed = NULL;
EventPipeEvent *EventPipeEventMonoProfilerVTableLoaded = NULL;
EventPipeEvent *EventPipeEventMonoProfilerModuleLoading = NULL;
EventPipeEvent *EventPipeEventMonoProfilerModuleFailed = NULL;
EventPipeEvent *EventPipeEventMonoProfilerModuleLoaded = NULL;
EventPipeEvent *EventPipeEventMonoProfilerModuleUnloading = NULL;
EventPipeEvent *EventPipeEventMonoProfilerModuleUnloaded = NULL;
EventPipeEvent *EventPipeEventMonoProfilerAssemblyLoading = NULL;
EventPipeEvent *EventPipeEventMonoProfilerAssemblyLoaded = NULL;
EventPipeEvent *EventPipeEventMonoProfilerAssemblyUnloading = NULL;
EventPipeEvent *EventPipeEventMonoProfilerAssemblyUnloaded = NULL;
EventPipeEvent *EventPipeEventMonoProfilerMethodEnter = NULL;
EventPipeEvent *EventPipeEventMonoProfilerMethodLeave = NULL;
EventPipeEvent *EventPipeEventMonoProfilerMethodTailCall = NULL;
EventPipeEvent *EventPipeEventMonoProfilerMethodExceptionLeave = NULL;
EventPipeEvent *EventPipeEventMonoProfilerMethodFree = NULL;
EventPipeEvent *EventPipeEventMonoProfilerMethodBeginInvoke = NULL;
EventPipeEvent *EventPipeEventMonoProfilerMethodEndInvoke = NULL;
EventPipeEvent *EventPipeEventMonoProfilerExceptionThrow = NULL;
EventPipeEvent *EventPipeEventMonoProfilerExceptionClause = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCEvent = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCAllocation = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCMoves = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCResize = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCHandleCreated = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCHandleDeleted = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCFinalizing = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCFinalized = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCFinalizingObject = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCFinalizedObject = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCRootRegister = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCRootUnregister = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCRoots = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCHeapDumpStart = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCHeapDumpStop = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCHeapDumpObjectReference = NULL;
EventPipeEvent *EventPipeEventMonoProfilerMonitorContention = NULL;
EventPipeEvent *EventPipeEventMonoProfilerMonitorFailed = NULL;
EventPipeEvent *EventPipeEventMonoProfilerMonitorAcquired = NULL;
EventPipeEvent *EventPipeEventMonoProfilerThreadStarted = NULL;
EventPipeEvent *EventPipeEventMonoProfilerThreadStopping = NULL;
EventPipeEvent *EventPipeEventMonoProfilerThreadStopped = NULL;
EventPipeEvent *EventPipeEventMonoProfilerThreadExited = NULL;
EventPipeEvent *EventPipeEventMonoProfilerThreadName = NULL;
EventPipeEvent *EventPipeEventMonoProfilerJitDoneVerbose = NULL;
EventPipeEvent *EventPipeEventMonoProfilerGCHeapDumpVTableClassReference = NULL;

bool EventPipeEventEnabledMonoProfilerContextLoaded(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerContextLoaded);
}

uint32_t EventPipeWriteEventMonoProfilerContextLoaded(
    const void* ObjectID,
    const uint64_t AppDomainID,
    const uint64_t ContextID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerContextLoaded())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)ObjectID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ContextID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerContextLoaded, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerContextUnloaded(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerContextUnloaded);
}

uint32_t EventPipeWriteEventMonoProfilerContextUnloaded(
    const void* ObjectID,
    const uint64_t AppDomainID,
    const uint64_t ContextID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerContextUnloaded())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)ObjectID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ContextID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerContextUnloaded, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerAppDomainLoading(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerAppDomainLoading);
}

uint32_t EventPipeWriteEventMonoProfilerAppDomainLoading(
    const uint64_t AppDomainID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerAppDomainLoading())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerAppDomainLoading, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerAppDomainLoaded(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerAppDomainLoaded);
}

uint32_t EventPipeWriteEventMonoProfilerAppDomainLoaded(
    const uint64_t AppDomainID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerAppDomainLoaded())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerAppDomainLoaded, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerAppDomainUnloading(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerAppDomainUnloading);
}

uint32_t EventPipeWriteEventMonoProfilerAppDomainUnloading(
    const uint64_t AppDomainID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerAppDomainUnloading())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerAppDomainUnloading, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerAppDomainUnloaded(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerAppDomainUnloaded);
}

uint32_t EventPipeWriteEventMonoProfilerAppDomainUnloaded(
    const uint64_t AppDomainID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerAppDomainUnloaded())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerAppDomainUnloaded, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerAppDomainName(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerAppDomainName);
}

uint32_t EventPipeWriteEventMonoProfilerAppDomainName(
    const uint64_t AppDomainID,
    const ep_char8_t* AppDomainName,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerAppDomainName())
        return ERROR_SUCCESS;

    size_t size = 72;
    uint8_t stackBuffer[72];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!AppDomainName) { AppDomainName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(AppDomainName, strlen((const char *)AppDomainName), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerAppDomainName, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerJitBegin(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerJitBegin);
}

uint32_t EventPipeWriteEventMonoProfilerJitBegin(
    const uint64_t MethodID,
    const uint64_t ModuleID,
    const uint32_t MethodToken,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerJitBegin())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(MethodToken, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerJitBegin, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerJitFailed(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerJitFailed);
}

uint32_t EventPipeWriteEventMonoProfilerJitFailed(
    const uint64_t MethodID,
    const uint64_t ModuleID,
    const uint32_t MethodToken,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerJitFailed())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(MethodToken, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerJitFailed, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerJitDone(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerJitDone);
}

uint32_t EventPipeWriteEventMonoProfilerJitDone(
    const uint64_t MethodID,
    const uint64_t ModuleID,
    const uint32_t MethodToken,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerJitDone())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(MethodToken, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerJitDone, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerJitDone_V1(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerJitDone_V1);
}

uint32_t EventPipeWriteEventMonoProfilerJitDone_V1(
    const uint64_t MethodID,
    const uint64_t ModuleID,
    const uint32_t MethodToken,
    const uint32_t Count,
    int Values_ElementSize,
    const void* Values,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerJitDone_V1())
        return ERROR_SUCCESS;

    size_t size = 56;
    uint8_t stackBuffer[56];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(MethodToken, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)Values, (int)Values_ElementSize * (int)Count, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerJitDone_V1, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerJitChunkCreated(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerJitChunkCreated);
}

uint32_t EventPipeWriteEventMonoProfilerJitChunkCreated(
    const void* ChunkID,
    const uint64_t ChunkSize,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerJitChunkCreated())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)ChunkID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ChunkSize, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerJitChunkCreated, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerJitChunkDestroyed(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerJitChunkDestroyed);
}

uint32_t EventPipeWriteEventMonoProfilerJitChunkDestroyed(
    const void* ChunkID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerJitChunkDestroyed())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)ChunkID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerJitChunkDestroyed, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerJitCodeBuffer(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerJitCodeBuffer);
}

uint32_t EventPipeWriteEventMonoProfilerJitCodeBuffer(
    const void* BufferID,
    const uint64_t BufferSize,
    const uint8_t BufferType,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerJitCodeBuffer())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)BufferID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(BufferSize, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint8_t(BufferType, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerJitCodeBuffer, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerClassLoading(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerClassLoading);
}

uint32_t EventPipeWriteEventMonoProfilerClassLoading(
    const uint64_t ClassID,
    const uint64_t ModuleID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerClassLoading())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(ClassID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerClassLoading, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerClassFailed(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerClassFailed);
}

uint32_t EventPipeWriteEventMonoProfilerClassFailed(
    const uint64_t ClassID,
    const uint64_t ModuleID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerClassFailed())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(ClassID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerClassFailed, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerClassLoaded(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerClassLoaded);
}

uint32_t EventPipeWriteEventMonoProfilerClassLoaded(
    const uint64_t ClassID,
    const uint64_t ModuleID,
    const ep_char8_t* ClassName,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerClassLoaded())
        return ERROR_SUCCESS;

    size_t size = 80;
    uint8_t stackBuffer[80];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!ClassName) { ClassName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(ClassID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ClassName, strlen((const char *)ClassName), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerClassLoaded, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerClassLoaded_V1(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerClassLoaded_V1);
}

uint32_t EventPipeWriteEventMonoProfilerClassLoaded_V1(
    const uint64_t ClassID,
    const uint64_t ModuleID,
    const ep_char8_t* ClassName,
    const uint32_t Count,
    int Values_ElementSize,
    const void* Values,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerClassLoaded_V1())
        return ERROR_SUCCESS;

    size_t size = 116;
    uint8_t stackBuffer[116];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!ClassName) { ClassName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(ClassID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ClassName, strlen((const char *)ClassName), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)Values, (int)Values_ElementSize * (int)Count, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerClassLoaded_V1, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerVTableLoading(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerVTableLoading);
}

uint32_t EventPipeWriteEventMonoProfilerVTableLoading(
    const uint64_t VTableID,
    const uint64_t ClassID,
    const uint64_t AppDomainID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerVTableLoading())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(VTableID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ClassID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerVTableLoading, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerVTableFailed(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerVTableFailed);
}

uint32_t EventPipeWriteEventMonoProfilerVTableFailed(
    const uint64_t VTableID,
    const uint64_t ClassID,
    const uint64_t AppDomainID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerVTableFailed())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(VTableID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ClassID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerVTableFailed, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerVTableLoaded(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerVTableLoaded);
}

uint32_t EventPipeWriteEventMonoProfilerVTableLoaded(
    const uint64_t VTableID,
    const uint64_t ClassID,
    const uint64_t AppDomainID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerVTableLoaded())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(VTableID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ClassID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerVTableLoaded, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerModuleLoading(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerModuleLoading);
}

uint32_t EventPipeWriteEventMonoProfilerModuleLoading(
    const uint64_t ModuleID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerModuleLoading())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerModuleLoading, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerModuleFailed(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerModuleFailed);
}

uint32_t EventPipeWriteEventMonoProfilerModuleFailed(
    const uint64_t ModuleID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerModuleFailed())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerModuleFailed, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerModuleLoaded(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerModuleLoaded);
}

uint32_t EventPipeWriteEventMonoProfilerModuleLoaded(
    const uint64_t ModuleID,
    const ep_char8_t* ModuleName,
    const ep_char8_t* ModuleSignature,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerModuleLoaded())
        return ERROR_SUCCESS;

    size_t size = 136;
    uint8_t stackBuffer[136];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!ModuleName) { ModuleName = (const ep_char8_t *)"NULL"; }
    if (!ModuleSignature) { ModuleSignature = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ModuleName, strlen((const char *)ModuleName), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ModuleSignature, strlen((const char *)ModuleSignature), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerModuleLoaded, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerModuleUnloading(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerModuleUnloading);
}

uint32_t EventPipeWriteEventMonoProfilerModuleUnloading(
    const uint64_t ModuleID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerModuleUnloading())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerModuleUnloading, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerModuleUnloaded(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerModuleUnloaded);
}

uint32_t EventPipeWriteEventMonoProfilerModuleUnloaded(
    const uint64_t ModuleID,
    const ep_char8_t* ModuleName,
    const ep_char8_t* ModuleSignature,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerModuleUnloaded())
        return ERROR_SUCCESS;

    size_t size = 136;
    uint8_t stackBuffer[136];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!ModuleName) { ModuleName = (const ep_char8_t *)"NULL"; }
    if (!ModuleSignature) { ModuleSignature = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ModuleName, strlen((const char *)ModuleName), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ModuleSignature, strlen((const char *)ModuleSignature), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerModuleUnloaded, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerAssemblyLoading(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerAssemblyLoading);
}

uint32_t EventPipeWriteEventMonoProfilerAssemblyLoading(
    const uint64_t AssemblyID,
    const uint64_t ModuleID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerAssemblyLoading())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(AssemblyID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerAssemblyLoading, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerAssemblyLoaded(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerAssemblyLoaded);
}

uint32_t EventPipeWriteEventMonoProfilerAssemblyLoaded(
    const uint64_t AssemblyID,
    const uint64_t ModuleID,
    const ep_char8_t* AssemblyName,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerAssemblyLoaded())
        return ERROR_SUCCESS;

    size_t size = 80;
    uint8_t stackBuffer[80];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!AssemblyName) { AssemblyName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(AssemblyID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(AssemblyName, strlen((const char *)AssemblyName), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerAssemblyLoaded, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerAssemblyUnloading(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerAssemblyUnloading);
}

uint32_t EventPipeWriteEventMonoProfilerAssemblyUnloading(
    const uint64_t AssemblyID,
    const uint64_t ModuleID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerAssemblyUnloading())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(AssemblyID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerAssemblyUnloading, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerAssemblyUnloaded(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerAssemblyUnloaded);
}

uint32_t EventPipeWriteEventMonoProfilerAssemblyUnloaded(
    const uint64_t AssemblyID,
    const uint64_t ModuleID,
    const ep_char8_t* AssemblyName,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerAssemblyUnloaded())
        return ERROR_SUCCESS;

    size_t size = 80;
    uint8_t stackBuffer[80];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!AssemblyName) { AssemblyName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(AssemblyID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(AssemblyName, strlen((const char *)AssemblyName), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerAssemblyUnloaded, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerMethodEnter(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerMethodEnter);
}

uint32_t EventPipeWriteEventMonoProfilerMethodEnter(
    const uint64_t MethodID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerMethodEnter())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerMethodEnter, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerMethodLeave(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerMethodLeave);
}

uint32_t EventPipeWriteEventMonoProfilerMethodLeave(
    const uint64_t MethodID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerMethodLeave())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerMethodLeave, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerMethodTailCall(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerMethodTailCall);
}

uint32_t EventPipeWriteEventMonoProfilerMethodTailCall(
    const uint64_t MethodID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerMethodTailCall())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerMethodTailCall, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerMethodExceptionLeave(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerMethodExceptionLeave);
}

uint32_t EventPipeWriteEventMonoProfilerMethodExceptionLeave(
    const uint64_t MethodID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerMethodExceptionLeave())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerMethodExceptionLeave, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerMethodFree(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerMethodFree);
}

uint32_t EventPipeWriteEventMonoProfilerMethodFree(
    const uint64_t MethodID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerMethodFree())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerMethodFree, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerMethodBeginInvoke(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerMethodBeginInvoke);
}

uint32_t EventPipeWriteEventMonoProfilerMethodBeginInvoke(
    const uint64_t MethodID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerMethodBeginInvoke())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerMethodBeginInvoke, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerMethodEndInvoke(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerMethodEndInvoke);
}

uint32_t EventPipeWriteEventMonoProfilerMethodEndInvoke(
    const uint64_t MethodID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerMethodEndInvoke())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerMethodEndInvoke, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerExceptionThrow(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerExceptionThrow);
}

uint32_t EventPipeWriteEventMonoProfilerExceptionThrow(
    const uint64_t TypeID,
    const void* ObjectID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerExceptionThrow())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(TypeID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uintptr_t((uintptr_t)ObjectID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerExceptionThrow, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerExceptionClause(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerExceptionClause);
}

uint32_t EventPipeWriteEventMonoProfilerExceptionClause(
    const uint8_t ClauseType,
    const uint32_t ClauseIdx,
    const uint64_t MethodID,
    const uint64_t TypeID,
    const void* ObjectID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerExceptionClause())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint8_t(ClauseType, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(ClauseIdx, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(TypeID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uintptr_t((uintptr_t)ObjectID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerExceptionClause, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCEvent(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCEvent);
}

uint32_t EventPipeWriteEventMonoProfilerGCEvent(
    const uint8_t GCEventType,
    const uint32_t GCGeneration,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCEvent())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint8_t(GCEventType, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(GCGeneration, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCEvent, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCAllocation(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCAllocation);
}

uint32_t EventPipeWriteEventMonoProfilerGCAllocation(
    const uint64_t VTableID,
    const void* ObjectID,
    const uint64_t ObjectSize,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCAllocation())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(VTableID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uintptr_t((uintptr_t)ObjectID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ObjectSize, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCAllocation, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCMoves(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCMoves);
}

uint32_t EventPipeWriteEventMonoProfilerGCMoves(
    const uint32_t Count,
    int Values_ElementSize,
    const void* Values,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCMoves())
        return ERROR_SUCCESS;

    size_t size = 36;
    uint8_t stackBuffer[36];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)Values, (int)Values_ElementSize * (int)Count, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCMoves, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCResize(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCResize);
}

uint32_t EventPipeWriteEventMonoProfilerGCResize(
    const uint64_t NewSize,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCResize())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(NewSize, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCResize, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCHandleCreated(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCHandleCreated);
}

uint32_t EventPipeWriteEventMonoProfilerGCHandleCreated(
    const uint64_t HandleID,
    const uint8_t HandleType,
    const void* ObjectID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCHandleCreated())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(HandleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint8_t(HandleType, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uintptr_t((uintptr_t)ObjectID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCHandleCreated, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCHandleDeleted(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCHandleDeleted);
}

uint32_t EventPipeWriteEventMonoProfilerGCHandleDeleted(
    const uint64_t HandleID,
    const uint8_t HandleType,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCHandleDeleted())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(HandleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint8_t(HandleType, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCHandleDeleted, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCFinalizing(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCFinalizing);
}

uint32_t EventPipeWriteEventMonoProfilerGCFinalizing(    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCFinalizing())
        return ERROR_SUCCESS;
    ep_write_event (EventPipeEventMonoProfilerGCFinalizing, (uint8_t *) NULL, 0, (const uint8_t *)ActivityId, (const uint8_t *)RelatedActivityId);

    return ERROR_SUCCESS;
}

bool EventPipeEventEnabledMonoProfilerGCFinalized(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCFinalized);
}

uint32_t EventPipeWriteEventMonoProfilerGCFinalized(    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCFinalized())
        return ERROR_SUCCESS;
    ep_write_event (EventPipeEventMonoProfilerGCFinalized, (uint8_t *) NULL, 0, (const uint8_t *)ActivityId, (const uint8_t *)RelatedActivityId);

    return ERROR_SUCCESS;
}

bool EventPipeEventEnabledMonoProfilerGCFinalizingObject(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCFinalizingObject);
}

uint32_t EventPipeWriteEventMonoProfilerGCFinalizingObject(
    const void* ObjectID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCFinalizingObject())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)ObjectID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCFinalizingObject, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCFinalizedObject(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCFinalizedObject);
}

uint32_t EventPipeWriteEventMonoProfilerGCFinalizedObject(
    const void* ObjectID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCFinalizedObject())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)ObjectID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCFinalizedObject, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCRootRegister(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCRootRegister);
}

uint32_t EventPipeWriteEventMonoProfilerGCRootRegister(
    const void* RootID,
    const uint64_t RootSize,
    const uint8_t RootType,
    const uint64_t RootKeyID,
    const ep_char8_t* RootKeyName,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCRootRegister())
        return ERROR_SUCCESS;

    size_t size = 89;
    uint8_t stackBuffer[89];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!RootKeyName) { RootKeyName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uintptr_t((uintptr_t)RootID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(RootSize, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint8_t(RootType, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(RootKeyID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(RootKeyName, strlen((const char *)RootKeyName), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCRootRegister, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCRootUnregister(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCRootUnregister);
}

uint32_t EventPipeWriteEventMonoProfilerGCRootUnregister(
    const void* RootID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCRootUnregister())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)RootID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCRootUnregister, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCRoots(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCRoots);
}

uint32_t EventPipeWriteEventMonoProfilerGCRoots(
    const uint32_t Count,
    int Values_ElementSize,
    const void* Values,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCRoots())
        return ERROR_SUCCESS;

    size_t size = 36;
    uint8_t stackBuffer[36];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)Values, (int)Values_ElementSize * (int)Count, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCRoots, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCHeapDumpStart(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCHeapDumpStart);
}

uint32_t EventPipeWriteEventMonoProfilerGCHeapDumpStart(
    const ep_char8_t* HeapCollectParam,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCHeapDumpStart())
        return ERROR_SUCCESS;

    size_t size = 64;
    uint8_t stackBuffer[64];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!HeapCollectParam) { HeapCollectParam = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_string_utf8_to_utf16_t(HeapCollectParam, strlen((const char *)HeapCollectParam), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCHeapDumpStart, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCHeapDumpStop(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCHeapDumpStop);
}

uint32_t EventPipeWriteEventMonoProfilerGCHeapDumpStop(    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCHeapDumpStop())
        return ERROR_SUCCESS;
    ep_write_event (EventPipeEventMonoProfilerGCHeapDumpStop, (uint8_t *) NULL, 0, (const uint8_t *)ActivityId, (const uint8_t *)RelatedActivityId);

    return ERROR_SUCCESS;
}

bool EventPipeEventEnabledMonoProfilerGCHeapDumpObjectReference(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCHeapDumpObjectReference);
}

uint32_t EventPipeWriteEventMonoProfilerGCHeapDumpObjectReference(
    const void* ObjectID,
    const uint64_t VTableID,
    const uint64_t ObjectSize,
    const uint8_t ObjectGeneration,
    const uint32_t Count,
    int Values_ElementSize,
    const void* Values,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCHeapDumpObjectReference())
        return ERROR_SUCCESS;

    size_t size = 61;
    uint8_t stackBuffer[61];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)ObjectID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(VTableID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ObjectSize, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint8_t(ObjectGeneration, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)Values, (int)Values_ElementSize * (int)Count, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCHeapDumpObjectReference, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerMonitorContention(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerMonitorContention);
}

uint32_t EventPipeWriteEventMonoProfilerMonitorContention(
    const void* ObjectID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerMonitorContention())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)ObjectID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerMonitorContention, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerMonitorFailed(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerMonitorFailed);
}

uint32_t EventPipeWriteEventMonoProfilerMonitorFailed(
    const void* ObjectID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerMonitorFailed())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)ObjectID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerMonitorFailed, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerMonitorAcquired(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerMonitorAcquired);
}

uint32_t EventPipeWriteEventMonoProfilerMonitorAcquired(
    const void* ObjectID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerMonitorAcquired())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)ObjectID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerMonitorAcquired, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerThreadStarted(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerThreadStarted);
}

uint32_t EventPipeWriteEventMonoProfilerThreadStarted(
    const uint64_t ThreadID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerThreadStarted())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(ThreadID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerThreadStarted, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerThreadStopping(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerThreadStopping);
}

uint32_t EventPipeWriteEventMonoProfilerThreadStopping(
    const uint64_t ThreadID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerThreadStopping())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(ThreadID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerThreadStopping, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerThreadStopped(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerThreadStopped);
}

uint32_t EventPipeWriteEventMonoProfilerThreadStopped(
    const uint64_t ThreadID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerThreadStopped())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(ThreadID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerThreadStopped, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerThreadExited(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerThreadExited);
}

uint32_t EventPipeWriteEventMonoProfilerThreadExited(
    const uint64_t ThreadID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerThreadExited())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(ThreadID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerThreadExited, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerThreadName(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerThreadName);
}

uint32_t EventPipeWriteEventMonoProfilerThreadName(
    const uint64_t ThreadID,
    const ep_char8_t* ThreadName,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerThreadName())
        return ERROR_SUCCESS;

    size_t size = 72;
    uint8_t stackBuffer[72];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!ThreadName) { ThreadName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(ThreadID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ThreadName, strlen((const char *)ThreadName), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerThreadName, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerJitDoneVerbose(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerJitDoneVerbose);
}

uint32_t EventPipeWriteEventMonoProfilerJitDoneVerbose(
    const uint64_t MethodID,
    const ep_char8_t* MethodNamespace,
    const ep_char8_t* MethodName,
    const ep_char8_t* MethodSignature,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerJitDoneVerbose())
        return ERROR_SUCCESS;

    size_t size = 200;
    uint8_t stackBuffer[200];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!MethodNamespace) { MethodNamespace = (const ep_char8_t *)"NULL"; }
    if (!MethodName) { MethodName = (const ep_char8_t *)"NULL"; }
    if (!MethodSignature) { MethodSignature = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(MethodNamespace, strlen((const char *)MethodNamespace), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(MethodName, strlen((const char *)MethodName), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(MethodSignature, strlen((const char *)MethodSignature), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerJitDoneVerbose, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMonoProfilerGCHeapDumpVTableClassReference(void)
{
    return ep_event_is_enabled(EventPipeEventMonoProfilerGCHeapDumpVTableClassReference);
}

uint32_t EventPipeWriteEventMonoProfilerGCHeapDumpVTableClassReference(
    const uint64_t VTableID,
    const uint64_t ClassID,
    const uint64_t ModuleID,
    const ep_char8_t* ClassName,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMonoProfilerGCHeapDumpVTableClassReference())
        return ERROR_SUCCESS;

    size_t size = 88;
    uint8_t stackBuffer[88];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!ClassName) { ClassName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(VTableID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ClassID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ClassName, strlen((const char *)ClassName), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMonoProfilerGCHeapDumpVTableClassReference, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

void EP_CALLBACK_CALLTYPE EventPipeEtwCallbackDotNETRuntimeMonoProfiler(const uint8_t *, unsigned long, uint8_t, uint64_t, uint64_t, EventFilterDescriptor *, void *);

void InitDotNETRuntimeMonoProfiler(void);

void InitDotNETRuntimeMonoProfiler(void)
{
    EventPipeProviderDotNETRuntimeMonoProfiler = create_provider(DotNETRuntimeMonoProfilerName, (EventPipeEtwCallbackDotNETRuntimeMonoProfiler));
    EventPipeEventMonoProfilerContextLoaded = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,1,8,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerContextUnloaded = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,2,8,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerAppDomainLoading = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,3,8,0,EP_EVENT_LEVEL_VERBOSE,false);
    EventPipeEventMonoProfilerAppDomainLoaded = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,4,8,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerAppDomainUnloading = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,5,8,0,EP_EVENT_LEVEL_VERBOSE,false);
    EventPipeEventMonoProfilerAppDomainUnloaded = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,6,8,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerAppDomainName = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,7,8,0,EP_EVENT_LEVEL_VERBOSE,false);
    EventPipeEventMonoProfilerJitBegin = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,8,16,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerJitFailed = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,9,16,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerJitDone = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,10,16,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerJitDone_V1 = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,10,16,1,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerJitChunkCreated = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,11,16,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerJitChunkDestroyed = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,12,16,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerJitCodeBuffer = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,13,16,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerClassLoading = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,14,549755813888,0,EP_EVENT_LEVEL_VERBOSE,false);
    EventPipeEventMonoProfilerClassFailed = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,15,549755813888,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerClassLoaded = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,16,549755813888,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerClassLoaded_V1 = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,16,549755813888,1,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerVTableLoading = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,17,549755813888,0,EP_EVENT_LEVEL_VERBOSE,false);
    EventPipeEventMonoProfilerVTableFailed = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,18,549755813888,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerVTableLoaded = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,19,549755813888,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerModuleLoading = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,20,8,0,EP_EVENT_LEVEL_VERBOSE,false);
    EventPipeEventMonoProfilerModuleFailed = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,21,8,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerModuleLoaded = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,22,8,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerModuleUnloading = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,23,8,0,EP_EVENT_LEVEL_VERBOSE,false);
    EventPipeEventMonoProfilerModuleUnloaded = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,24,8,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerAssemblyLoading = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,25,8,0,EP_EVENT_LEVEL_VERBOSE,false);
    EventPipeEventMonoProfilerAssemblyLoaded = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,26,8,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerAssemblyUnloading = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,27,8,0,EP_EVENT_LEVEL_VERBOSE,false);
    EventPipeEventMonoProfilerAssemblyUnloaded = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,28,8,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerMethodEnter = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,29,536870912,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerMethodLeave = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,30,536870912,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerMethodTailCall = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,31,536870912,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerMethodExceptionLeave = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,32,536870912,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerMethodFree = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,33,536870912,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerMethodBeginInvoke = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,34,536870912,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerMethodEndInvoke = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,35,536870912,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerExceptionThrow = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,36,32768,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventMonoProfilerExceptionClause = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,37,32768,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerGCEvent = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,38,1,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerGCAllocation = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,39,2097152,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventMonoProfilerGCMoves = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,40,4194304,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerGCResize = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,41,33554432,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerGCHandleCreated = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,42,2,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventMonoProfilerGCHandleDeleted = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,43,2,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventMonoProfilerGCFinalizing = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,44,16777216,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerGCFinalized = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,45,16777216,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerGCFinalizingObject = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,46,16777216,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerGCFinalizedObject = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,47,16777216,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerGCRootRegister = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,48,67108864,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerGCRootUnregister = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,49,67108864,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerGCRoots = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,50,67108864,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerGCHeapDumpStart = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,51,1048576,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerGCHeapDumpStop = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,52,1048576,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerGCHeapDumpObjectReference = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,53,1048576,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerMonitorContention = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,54,1099511644160,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventMonoProfilerMonitorFailed = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,55,1099511627776,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventMonoProfilerMonitorAcquired = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,56,1099511627776,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventMonoProfilerThreadStarted = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,57,65536,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerThreadStopping = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,58,65536,0,EP_EVENT_LEVEL_VERBOSE,false);
    EventPipeEventMonoProfilerThreadStopped = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,59,65536,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerThreadExited = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,60,65536,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMonoProfilerThreadName = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,61,65536,0,EP_EVENT_LEVEL_VERBOSE,false);
    EventPipeEventMonoProfilerJitDoneVerbose = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,62,16,0,EP_EVENT_LEVEL_VERBOSE,false);
    EventPipeEventMonoProfilerGCHeapDumpVTableClassReference = provider_add_event(EventPipeProviderDotNETRuntimeMonoProfiler,63,134217728,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
}
#endif
